#-------------------------------------------------------------------------------
#
# CavernosaBlender.py
#
# Copyright (C) 2000 by Andrew H. Cox. All rights reserved.
#
# ToDo:
# > Look into explictly unloading Cavernosa on exit.
#
# Note:
# To Add a button:
#   1: add a button event handler
#      (make sure any globals it writes to are listed in func globals)
#   2: add a symbolic name for the label to the constants at the top of the file.
#   3: add a symbolic name for the button event to the events at the top of the file.

"""Copyright (C) 2000 by Andrew H. Cox. All rights reserved.

Reach me at: acox@globalnet.co.uk"""

import Blender
from Blender.Draw import *
from Blender.BGL import *
from Blender import NMesh
import Cavernosa

#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------
# Constants:
MAX_CAVERNOSA_FILENAME = Cavernosa.query_int(Cavernosa.IntMaxFixedStringSize)
MAX_BLENDER_NAME_LENGTH = 16 # Max length of names of objects in blender (conservative).
CHAR_WIDTH = 7
CHAR_HEIGHT = 12
MIN_BUTTON_HEIGHT = CHAR_HEIGHT
MAX_BUTTON_HEIGHT = 50
LABEL_X_BORDER = 10 # The combined space to the left and right of a button label.
MIN_DISPLAY_BOX_DIMENSION = 8
MAX_DISPLAY_BOX_DIMENSION = 999
MIN_EXPORT_BOX_DST_DIMENSION = 0.5
MAX_EXPORT_BOX_DST_DIMENSION = 32
MIN_GRID_LEVELS = 3
MAX_GRID_LEVELS = 8

#-------------------------------------------------------------------------------
# Button Events:
event_count = 1
SP_B_EVENT_WIDTH = event_count
event_count = event_count + 1
SP_B_EVENT_B_HEIGHT = event_count
event_count = event_count + 1
#Display events:
SP_B_EVENT_FOG_TOG = event_count
event_count = event_count + 1
SP_B_EVENT_WIRE_TOG = event_count
event_count = event_count + 1
SP_B_EVENT_DBSTICK_TOG = event_count
event_count = event_count + 1
SP_B_EVENT_COUNT_FACES_TOG = event_count
event_count = event_count + 1
SP_B_EVENT_DISP_BOX_X = event_count
event_count = event_count + 1
SP_B_EVENT_DISP_BOX_Y = event_count
event_count = event_count + 1
SP_B_EVENT_DISP_BOX_Z = event_count
event_count = event_count + 1
SP_B_EVENT_TWIST_COMP_TOG = event_count
event_count = event_count + 1
SP_B_EVENT_CREDIT_TOG = event_count
event_count = event_count + 1
SP_B_EVENT_RENDER_GRID_TOG = event_count
event_count = event_count + 1
# Control Events:
SP_B_EVENT_SENSITIVITY = event_count
event_count = event_count + 1
SP_B_EVENT_MOVEMENT = event_count
event_count = event_count + 1
SP_B_EVENT_CHANGE_AFFECTOR = event_count
event_count = event_count + 1
#Export events:
SP_B_EVENT_EXPORT_NAME = event_count
event_count = event_count + 1
SP_B_EVENT_EXPORT_TO_NAMED = event_count
event_count = event_count + 1
#SP_B_EVENT_EXPORT_TO_SELECTED = event_count
#event_count = event_count + 1
SP_B_EVENT_EXPORT_DIMENSION = event_count
event_count = event_count + 1
# Load/Save events:
SP_B_EVENT_SAVE_NAME = event_count
event_count = event_count + 1
SP_B_EVENT_SAVE = event_count
event_count = event_count + 1
SP_B_EVENT_LOAD = event_count
event_count = event_count + 1
SP_B_EVENT_NEW_GRID = event_count
event_count = event_count + 1
SP_B_EVENT_GRID_LEVELS = event_count
event_count = event_count + 1
SP_B_EVENT_QUIT = event_count

#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------
# Globals:
g_mouse_x = 255
g_mouse_y = 255
g_scissor_box = Buffer(GL_FLOAT, 4)

#-------------------------------------------------------------------------------
# GUI:
g_title_text = 'Cavernosa'
g_author_text = 'acox@globalnet.co.uk' # Remove this line !
g_button_height = 20
g_text_height = CHAR_HEIGHT
g_text_spacing = 5
g_text_offset = 5


#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------
# Side Panel:

#-------------------------------------------------------------------------------
# Widget Labels:
spl_width = 'PW: '
spl_b_height = 'BH: '

spl_fog_tog = 'F'
if Cavernosa.query_bool(Cavernosa.BoolWireGrid) == 1: spl_wire_tog = 'S'
else: spl_wire_tog = 'W'
spl_db_stick_tog = 'DB'
spl_render_grid_tog = 'R'
spl_count_faces_tog = 'Count Faces'
spl_disp_box_x_num = 'x: '
spl_disp_box_y_num = 'y: '
spl_disp_box_z_num = 'z: '
spl_twist_comp_tog = 'T'
spl_credit_tog = 'C'

spl_export_dim = 'Dimension: '
spl_export_name = 'Mesh: '
spl_export_to_named = 'To Mesh'

spl_save_name = 'File: '

#-------------------------------------------------------------------------------
# Side Panel Globals:
# FixMe: Load default toggle state from Cavernosa dll on script start.
# If they fix the bug I can add a button to the gui that sets this to 0:
g_enable_button_event_fix = 1
g_side_panel_width = 156

g_fog_tog = Cavernosa.query_bool(Cavernosa.BoolFoggingOn)
g_wire_tog = Cavernosa.query_bool(Cavernosa.BoolWireGrid)
g_twist_comp_tog = Cavernosa.query_bool(Cavernosa.BoolCompensateForTwist)
g_credit_tog = 1
g_render_grid_tog = 1
g_disp_box_x_num = Cavernosa.query_int(Cavernosa.IntDisplayBoxX)
g_disp_box_y_num = Cavernosa.query_int(Cavernosa.IntDisplayBoxY)
g_disp_box_z_num = Cavernosa.query_int(Cavernosa.IntDisplayBoxZ)
g_count_faces_tog = Cavernosa.query_bool(Cavernosa.BoolCountFaces)

g_sensitivity = 1.0
g_db_stick_tog = Cavernosa.query_bool(Cavernosa.BoolDisplayBoxStickOn)
g_motion_is_proportional = 1

g_export_dim = Cavernosa.query_float(Cavernosa.FloatExportBoxMaxDim)
g_export_name = 'Cavernosa'

g_save_name = 'Test.vox'
g_grid_levels = Cavernosa.query_int(Cavernosa.IntGridLevels)

#-------------------------------------------------------------------------------
def display():
    global g_pane_x, g_pane_y, g_pane_width, g_pane_height, g_side_panel_left
    try:

        # print 'Cavernosa Script display func called'
        # Find the screen box available for rendering into:
        glGetFloatv(GL_SCISSOR_BOX, g_scissor_box)
        g_pane_x = g_scissor_box[0]
        g_pane_y = g_scissor_box[1]
        g_pane_width = g_scissor_box[2]
        g_pane_height = g_scissor_box[3]
        g_side_panel_left = g_pane_width - g_side_panel_width # In pane relative coords.

        # Main 3d view:
        try:
            error = glGetError()
            glPushAttrib(GL_ALL_ATTRIB_BITS)
            glMatrixMode(GL_MODELVIEW)
            glPushMatrix()
            glMatrixMode(GL_PROJECTION)
            glPushMatrix()
            #glPushClientAttrib(GL_ALL_CLIENT_ATTRIB_BITS)
            
            glScissor(g_pane_x, g_pane_y, g_side_panel_left, g_pane_height)
            Cavernosa.reshape(g_pane_x, g_pane_y, g_side_panel_left, g_pane_height)
            if g_render_grid_tog == 1:
                Cavernosa.display()

        finally:
            new_error = glGetError()
            if error != new_error: print "glError, old: ", error, ", new: ", new_error

            #glPopClientAttrib()
            glMatrixMode(GL_PROJECTION)
            glPopMatrix()
            glMatrixMode(GL_MODELVIEW)
            glPopMatrix()
            glPopAttrib()
        
        glColor3f(1,1,1) # FixMe: this affects rendering of whole side panel: give spacer a text colour member.
            
        # Draw the Author Credit:
        if g_credit_tog == 1:
            draw_text(10, 10, 'Created by Andrew H. Cox: acox@globalnet.co.uk')

        # Draw the side panel:
        # ! Scissor box doesn't affect Draw gui widgets, just the glClear().
        glScissor(g_pane_x + g_side_panel_left, g_pane_y, g_side_panel_width, g_pane_height)
        glClearColor(0.5, 0.5, 0.5, 0.0)
        glClear(GL_COLOR_BUFFER_BIT)
        # Create the side panel, storing refrences to the widgets into the global scope's dictionary:
        draw_side_panel(g_side_panel_left, 0, g_side_panel_width, g_pane_height, g_pane_width, globals())

        if g_render_grid_tog == 1:
            Redraw(1)
    except:
        print 'Exception caught in display().  <<<<<<<<<<<<<<------------------------------<<<<<<<<<<<<<< XXXXX'


#-------------------------------------------------------------------------------
# The widget bag is a dictionary-like object.
# That includes the global dictionary to have the function assign to globals.
def draw_side_panel(x, y, width, height, pane_width, wb):
    
    s = VerticalSpacer(x, y, width, height)
    s.text('Cavernosa')
    #s.text('')
    
    s.set_text_alignment(s.TEXT_LEFT_ALIGNED)
    
    s.text('Resize this:')
    wb['spw_width'] = s.slider(spl_width, SP_B_EVENT_WIDTH, width, pane_width * 0.07, pane_width * 0.5, 1, 'Width of side panel')
    wb['spw_b_height'] = s.slider(spl_b_height, SP_B_EVENT_B_HEIGHT, g_button_height, MIN_BUTTON_HEIGHT, MAX_BUTTON_HEIGHT, 1, 'Height of side panel buttons')
    
    s.text('Display:')
    s.begin_horiz_packing()
    #Fog:
    wb['spw_fog_tog'] = s.horiz_packed_toggle(spl_fog_tog, SP_B_EVENT_FOG_TOG, g_fog_tog, "Toggle Fogging On/Off")
    # Wireframe/Solid grid:
    wb['spw_wire_tog'] = s.horiz_packed_toggle(spl_wire_tog, SP_B_EVENT_WIRE_TOG, g_wire_tog, "Toggle Wireframe/Solid grid")
    #DisplayBox Stick to current position:
    wb['spw_db_stick_tog'] = s.horiz_packed_toggle(spl_db_stick_tog, SP_B_EVENT_DBSTICK_TOG, g_db_stick_tog, "Toggle Display box follows focus On/Off")
    #Camera Twist compensation:
    wb['spw_twist_comp_tog'] = s.horiz_packed_toggle(spl_twist_comp_tog, SP_B_EVENT_TWIST_COMP_TOG, g_twist_comp_tog, "Toggle Camera twist compensation On/Off")
    #Author credit:
    wb['spw_credit_tog'] = s.horiz_packed_toggle(spl_credit_tog, SP_B_EVENT_CREDIT_TOG, g_credit_tog, "Toggle Author credit On/Off")
    #Rendering the grid:
    wb['spw_render_grid_tog'] = s.horiz_packed_toggle(spl_render_grid_tog, SP_B_EVENT_RENDER_GRID_TOG, g_render_grid_tog, "Toggle Grid Rendering On/Off")
    s.end_horiz_packing()
    wb['spw_count_faces_tog'] = s.toggle(spl_count_faces_tog, SP_B_EVENT_COUNT_FACES_TOG, g_count_faces_tog, "Count faces in display box and print total")

    s.text('Display Box:')
    s.begin_horiz_packing()
    #X,Y, Z display box dimensions:
    wb['spw_disp_box_x_num'] = s.horiz_packed_number(spl_disp_box_x_num, SP_B_EVENT_DISP_BOX_X, g_disp_box_x_num, MIN_DISPLAY_BOX_DIMENSION, MAX_DISPLAY_BOX_DIMENSION, 3, "Display Box x dimension")
    wb['spw_disp_box_y_num'] = s.horiz_packed_number(spl_disp_box_y_num, SP_B_EVENT_DISP_BOX_Y, g_disp_box_y_num, MIN_DISPLAY_BOX_DIMENSION, MAX_DISPLAY_BOX_DIMENSION, 3, "Display Box y dimension")
    wb['spw_disp_box_z_num'] = s.horiz_packed_number(spl_disp_box_z_num, SP_B_EVENT_DISP_BOX_Z, g_disp_box_z_num, MIN_DISPLAY_BOX_DIMENSION, MAX_DISPLAY_BOX_DIMENSION, 3, "Display Box z dimension")
    s.end_horiz_packing()

    s.text('Control:')
    wb['spw_sensitivity'] = s.slider('Sensitivity: ', SP_B_EVENT_SENSITIVITY, g_sensitivity, 0.1, 2, 1, 'Scaling of movements')
    if(g_motion_is_proportional == 1):
        movement = 'Proportional'
    else:
        movement = 'Direct'
    s.button(movement, SP_B_EVENT_MOVEMENT, "Change between proportional and direct control")
    s.button('Change Affector', SP_B_EVENT_CHANGE_AFFECTOR, "Change the current tool 'bit'")
        
    s.text('Export:')
    # Export Name String (FixMe):
    wb['spw_export_dim'] = s.number(spl_export_dim, SP_B_EVENT_EXPORT_DIMENSION, g_export_dim, MIN_EXPORT_BOX_DST_DIMENSION, MAX_EXPORT_BOX_DST_DIMENSION, "Max dimension of generated mesh")
    wb['spw_export_name'] = s.string(spl_export_name, SP_B_EVENT_EXPORT_NAME, g_export_name, MAX_BLENDER_NAME_LENGTH, 'The name of the object and mesh to export to')
    s.button(spl_export_to_named, SP_B_EVENT_EXPORT_TO_NAMED, "Export the Display Box contents to above mesh")
    #s.button('Export To Selected', SP_B_EVENT_EXPORT_TO_SELECTED, "Export the Display Box contents to above mesh and bind to selected object")
    
    s.text('Save/Load:')
    wb['spw_save_name'] = s.string(spl_save_name, SP_B_EVENT_SAVE_NAME, g_save_name, MAX_CAVERNOSA_FILENAME, 'The name of the file to save the voxel grid to')
    s.button('Save', SP_B_EVENT_SAVE, "Save the Display Box contents as a file named in the field above")
    s.button('Load', SP_B_EVENT_LOAD, "Load the Display Box contents from a file named above")
    s.button('New Grid', SP_B_EVENT_NEW_GRID, "Replaces the current grid with an empty new one")
    wb['spw_grid_levels_num'] = s.number('New Grid Levels: ', SP_B_EVENT_GRID_LEVELS, g_grid_levels, MIN_GRID_LEVELS, MAX_GRID_LEVELS, "Levels in heirarchical grid. 8 Levels == 16'777'216 cubed.")
        
    #s.space_for_button()
    s.button('Quit', SP_B_EVENT_QUIT, "Exit Cavernosa")
 
    #s.set_text_alignment(s.TEXT_CENTRED)   
    #s.text('')    
    #s.text_lines(['poop','scoop'])
    #s.text('123')
    #s.set_text_alignment(s.TEXT_LEFT_ALIGNED)
    #s.text('123')
    #s.set_text_alignment(s.TEXT_RIGHT_ALIGNED)
    #s.text('123')    
    #s.text('1234567890')
    #s.text('1234567890123456789')
    

#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------
def do_exit():
    print
    print 'Exiting'
    print
    # FixMe: un-import Cavernosa.
    Blender.Draw.Exit()


#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------
def event(evt, val):  
    global g_mouse_x  
    global g_mouse_y
    
    try:
        if (evt == MOUSEX): 
            g_mouse_x = val
            Cavernosa.motion(g_mouse_x, g_mouse_y)
        elif (evt == MOUSEY):
            g_mouse_y = val
            Cavernosa.motion(g_mouse_x, g_mouse_y)
        elif (evt == LEFTMOUSE):
            if (val == 0): val = 1
            else: val = 0
            Cavernosa.button(0,val)
        elif (evt == RIGHTMOUSE):
            if (val == 0): val = 1
            else: val = 0
            Cavernosa.button(2,val)
        elif (evt== VKEY): Cavernosa.key(ord('v'), val)
        elif (evt== FKEY): Cavernosa.key(ord('f'), val)
        elif (evt== CKEY): Cavernosa.key(ord('c'), val)
        elif (evt== DKEY): Cavernosa.key(ord('d'), val)
        elif (evt== SKEY): Cavernosa.key(ord('s'), val)
        elif (evt== XKEY): Cavernosa.key(ord('x'), val)
        elif (evt== ZKEY): Cavernosa.key(ord('z'), val)
        elif (evt== AKEY): Cavernosa.key(ord('a'), val)
        elif (evt== SPACEKEY): Cavernosa.key(ord(' '), val)
        elif (evt== BKEY): Cavernosa.key(ord('b'), val) # ! Not used
        elif (evt== GKEY): Cavernosa.key(ord('g'), val) # ! Not used
        #elif (evt== QKEY): do_exit()
        
        #Redraw(1)
    except:
        print 'Exception caught in event(). <<<<<<<<<<<<<<------------------------------<<<<<<<<<<<<<< XXXXX'


#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------
g_bevent_count = 0
def bevent(evt):
    global g_bevent_count
    global g_side_panel_width, g_button_height, g_fog_tog, g_export_name, g_credit_tog, g_db_stick_tog, g_save_name, g_render_grid_tog
    global g_wire_tog, spl_wire_tog, g_twist_comp_tog, g_count_faces_tog, g_disp_box_x_num, g_disp_box_y_num, g_disp_box_z_num
    global g_export_dim, g_sensitivity, g_motion_is_proportional, g_grid_levels
    
    g_bevent_count = g_bevent_count + 1
    #print 'Button event handler called ', g_bevent_count
    if (g_bevent_count & 1) != 1:
        return
        #if g_enable_button_event_fix == 1: return
    #print 'Button event handler function body entered.'
    try:
        #---------------
        # Display events:
        if evt == SP_B_EVENT_WIDTH:
            g_side_panel_width = spw_width.val
            #Redraw(0)

        elif evt == SP_B_EVENT_B_HEIGHT:
            g_button_height = spw_b_height.val
            
        # Fogging:
        elif evt == SP_B_EVENT_FOG_TOG:
            g_fog_tog = spw_fog_tog.val
            Cavernosa.set_bool(Cavernosa.BoolFoggingOn, g_fog_tog)

        # Wireframe/Solid grid:
        elif evt == SP_B_EVENT_WIRE_TOG:
            g_wire_tog = spw_wire_tog.val
            Cavernosa.set_bool(Cavernosa.BoolWireGrid, g_wire_tog)
            if g_wire_tog == 1: spl_wire_tog = 'W'
            else: spl_wire_tog = 'S'

            
        # Display box toggle between sticky and follow focus point:
        elif evt == SP_B_EVENT_DBSTICK_TOG:
            g_db_stick_tog = spw_db_stick_tog.val
            Cavernosa.do_action(Cavernosa.ActionToggleDisplayBoxFixed)
        
        # Author credit:
        elif evt == SP_B_EVENT_CREDIT_TOG:
            g_credit_tog = spw_credit_tog.val
            
        #Grid Rendering toggle:
        elif evt == SP_B_EVENT_RENDER_GRID_TOG:
            g_render_grid_tog = spw_render_grid_tog.val
            
        #Face counting toggle:
        elif evt == SP_B_EVENT_COUNT_FACES_TOG:
            g_count_faces_tog = spw_count_faces_tog.val
            Cavernosa.set_bool(Cavernosa.BoolCountFaces, g_count_faces_tog)
            
        # Display Box Dimensions:
        elif evt == SP_B_EVENT_DISP_BOX_X:
            g_disp_box_x_num = spw_disp_box_x_num.val
            Cavernosa.set_int(Cavernosa.IntDisplayBoxX, g_disp_box_x_num)
            Cavernosa.do_action(Cavernosa.ActionUpdateDisplayBox)

        elif evt == SP_B_EVENT_DISP_BOX_Y:
            g_disp_box_y_num = spw_disp_box_y_num.val
            Cavernosa.set_int(Cavernosa.IntDisplayBoxY, g_disp_box_y_num)
            Cavernosa.do_action(Cavernosa.ActionUpdateDisplayBox)

        elif evt == SP_B_EVENT_DISP_BOX_Z:
            g_disp_box_z_num = spw_disp_box_z_num.val
            Cavernosa.set_int(Cavernosa.IntDisplayBoxZ, g_disp_box_z_num)
            Cavernosa.do_action(Cavernosa.ActionUpdateDisplayBox)

        # ----------------
        # Control/Movement:
            
        # 
        elif evt == SP_B_EVENT_SENSITIVITY:
            g_sensitivity = spw_sensitivity.val
            Cavernosa.set_float(Cavernosa.FloatMovementScaling, g_sensitivity)

        # Camera twist compensation:
        elif evt == SP_B_EVENT_TWIST_COMP_TOG:
            g_twist_comp_tog = spw_twist_comp_tog.val
            Cavernosa.set_bool(Cavernosa.BoolCompensateForTwist, g_twist_comp_tog)
        
        # Direct/Proportional control toggle:
        elif evt == SP_B_EVENT_MOVEMENT:
            Cavernosa.do_action(Cavernosa.ActionToggleMovement)
            if g_motion_is_proportional == 0: g_motion_is_proportional = 1
            else: g_motion_is_proportional = 0

        elif evt == SP_B_EVENT_CHANGE_AFFECTOR:
            Cavernosa.do_action(Cavernosa.ActionChangeGridAffector)

        # ----------------
        # Exporting events:
            
        elif evt == SP_B_EVENT_EXPORT_DIMENSION:
            g_export_dim = spw_export_dim.val
            Cavernosa.set_float(Cavernosa.FloatExportBoxMaxDim, g_export_dim)

        elif evt == SP_B_EVENT_EXPORT_NAME:
            g_export_name = spw_export_name.val

        # Export to named object event:
        elif evt == SP_B_EVENT_EXPORT_TO_NAMED:
            export_to_named(g_export_name)
            
        #elif evt == SP_B_EVENT_EXPORT_TO_SELECTED:
            #export_to_selected(g_export_name)

        # ---------------------
        # Saving/Loading events:
        
        elif evt == SP_B_EVENT_SAVE_NAME:
            g_save_name = spw_save_name.val
        
        elif evt == SP_B_EVENT_SAVE:
            Cavernosa.set_string(Cavernosa.StringSavePath, g_save_name)
            Cavernosa.do_action(Cavernosa.ActionSave) # Load
        
        elif evt == SP_B_EVENT_LOAD:
            Cavernosa.set_string(Cavernosa.StringSavePath, g_save_name)
            Cavernosa.do_action(Cavernosa.ActionLoad) # Load

        elif evt == SP_B_EVENT_NEW_GRID:
            Cavernosa.do_action(Cavernosa.ActionClearGrid) # Clear

        elif evt == SP_B_EVENT_GRID_LEVELS:
            g_grid_levels = spw_grid_levels_num.val
            Cavernosa.set_int(Cavernosa.IntGridLevels, g_grid_levels)

        #----------
        # Exit/Quit:
        elif evt == SP_B_EVENT_QUIT:
            do_exit()
        
        Redraw(1)
    except:
        print 'Exception caught in bevent(). <<<<<<<<<<<<<<------------------------------<<<<<<<<<<<<<< XXXXX'

#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------
# Callbacks:


#-------------------------------------------------------------------------------
def vertex3f(x, y, z):
    global g_m, g_corner_count, g_face_corners

    v = NMesh.Vert(x, y, z)
    g_face_corners[g_corner_count] = v
    g_m.verts.append(v)

    if g_corner_count == 3:
        f = NMesh.Face()
        c = NMesh.Col() # White by default.
        f.v.append(g_face_corners[0])
        f.v.append(g_face_corners[1])
        f.v.append(g_face_corners[2])
        f.v.append(g_face_corners[3])
        f.col.append(c)
        f.col.append(c)
        f.col.append(c)
        f.col.append(c)
        
        # FixMe: Set Colour?
        g_m.faces.append(f)
        g_corner_count = 0
    else:
        g_corner_count = g_corner_count + 1
    #print 'vertex3f(x, y, z) called: ', x, y, z

#-------------------------------------------------------------------------------
def export_to_named(name):
    global g_m, g_corner_count, g_face_corners
    g_m = NMesh.GetRaw()
    
    g_corner_count = 0
    g_face_corners = [None,None,None,None]
    Cavernosa.set_vertex3_func(vertex3f)
    Cavernosa.do_action(Cavernosa.ActionExportDisplayBoxAsSquares) # Export
    
    # To put it back, get the object called name and add the mesh to it:(FixMe)
    g_m.has_uvco = 0
    g_m.has_col = 1
    NMesh.PutRaw(g_m, name)
    
#-------------------------------------------------------------------------------
# Export to raw mesh and bind the mesh to the selected object. (FixMe)
def export_to_selected(name):
    global g_m, g_corner_count, g_face_corners
    g_m = NMesh.GetRaw()
    
    g_corner_count = 0
    g_face_corners = [None,None,None,None]
    Cavernosa.set_vertex3_func(vertex3f)
    Cavernosa.do_action(Cavernosa.ActionExportDisplayBoxAsSquares) # Export
    
    # To put it back, get the object called name and add the mesh to it:(FixMe)
    g_m.has_uvco = 0
    g_m.has_col = 1
    NMesh.PutRaw(g_m, name)

#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------
# Utility functions:

# pane_to_window() transform a point in the script pane to the window's
# coordinate system:
def pane_x_to_window(x):
    return g_pane_x + x
def pane_y_to_window(y):
    return g_pane_y + y
    
def draw_text(x, y, text):
    glRasterPos2i(x,y)
    Text(text)

# Length of string in pixels:
def string_width(text):
    return len(text) * CHAR_WIDTH

#-------------------------------------------------------------------------------
# VerticalSpacer: as a colum of widgets is drawn flush, this tracks the current
# y coordinate to draw the next widget.
class VerticalSpacer:
    # Modes for aligned text printed into the widget stack:
    TEXT_CENTRED = 0
    TEXT_LEFT_ALIGNED = 1
    TEXT_RIGHT_ALIGNED = 2
    
    def __init__(s, x, y, width, height):
    	# !Should realy be passed all those globals rather than going and gettiing them itself.
        s.left = x
        s.width = width
        s.height = height
        s.base_y = y # More meaningful name for y when used by drawing members.
        s.top_offset = height # Current insertion point offset from base.
        s.text_height = g_text_height
        s.text_spacing = g_text_spacing
        s.text_alignment = s.TEXT_CENTRED
        s.text_edge_offset = g_text_offset
        s.button_height = g_button_height
    
    #-------------------------------------------------------------------------------
    #-------------------------------------------------------------------------------
    # Set methods to avoid typos creating unwanted data members:
    
    def set_text_alignment(s, v):
        s.text_alignment = v
    def set_text_edge_offset(s, v):
        s.text_edge_offset = v

    #-------------------------------------------------------------------------------
    #-------------------------------------------------------------------------------
    # Functions that draw elements of the stack (User callable functions):
        
    #-------------------------------------------------------------------------------
    def text(s, text):
        s.space_before_text()
        s.space_for_text()
        if s.text_alignment == s.TEXT_CENTRED:
            width = len(text) * CHAR_WIDTH
            pos = (s.left + (s.width / 2)) - (width / 2)
            draw_text(pos , s.y(), text)
        elif s.text_alignment == s.TEXT_LEFT_ALIGNED:
            draw_text(s.left + s.text_edge_offset, s.y(), text)
        elif s.text_alignment == s.TEXT_RIGHT_ALIGNED:
            width = len(text) * CHAR_WIDTH
            pos = (s.left + s.width) - (width + s.text_edge_offset)
            draw_text(pos , s.y(), text)
        s.space_after_text()

    #-------------------------------------------------------------------------------
    # Multiline version of text that doesn't double space between lines:
    def text_lines(s, lines):
        s.space_before_text()
        for text in lines:
            if s.text_alignment == s.TEXT_CENTRED:
                width = len(text) * CHAR_WIDTH
                pos = (s.left + (s.width / 2)) - (width / 2)
                draw_text(pos , s.y(), text)
            elif s.text_alignment == s.TEXT_LEFT_ALIGNED:
                draw_text(s.left + s.text_edge_offset, s.y(), text)
            elif s.text_alignment == s.TEXT_RIGHT_ALIGNED:
                width = len(text) * CHAR_WIDTH
                pos = (s.left + s.width) - (width + s.text_edge_offset)
                draw_text(pos , s.y(), text)
            s.space_for_text()
        s.space_after_text()
        
    #-------------------------------------------------------------------------------
    def button(s, label, event, tooltip):
        s.space_for_button()
        return Button(label, event, s.left, s.y(), s.width, s.button_height, tooltip)

    #-------------------------------------------------------------------------------
    def slider(s, label, event, initial, min, max, update, tooltip):
        s.space_for_button()
        return Slider(label, event, s.left, s.y(), s.width, s.button_height, initial, min, max, update, tooltip)

    #-------------------------------------------------------------------------------
    def string(s, label, event, initial, length, tooltip):
        s.space_for_button()
        return String(label, event, s.left, s.y(), s.width, s.button_height, initial, length, tooltip)

    #-------------------------------------------------------------------------------
    def toggle(s, label, event, default, tooltip):
        s.space_for_button()
        return Toggle(label, event, s.left, s.y(), s.width, s.button_height, default, tooltip)
        
    #-------------------------------------------------------------------------------
    def number(s, label, event, initial, min, max, tooltip):
        s.space_for_button()
        return Number(label, event, s.left, s.y(), s.width, s.button_height, initial, min, max, tooltip)
        
    #-------------------------------------------------------------------------------
    def begin_horiz_packing(s):
        s.space_for_button()
        s.x_offset = 0
    #-------------------------------------------------------------------------------
    def end_horiz_packing(s):
        pass

    #-------------------------------------------------------------------------------        
    def horiz_packed_toggle(s, label, event, default, tooltip):
        # Make space on row for toggle:
        width = string_width(label) + LABEL_X_BORDER
        s.horiz_space_for_button(width)        
        x = s.x() - width
        return Toggle(label, event, x, s.y(), width, s.button_height, default, tooltip)

    #-------------------------------------------------------------------------------
    def horiz_packed_number(s, label, event, initial, min, max, digits, tooltip):
        # Make space on row for button:
        width = string_width(label) + LABEL_X_BORDER + (digits * CHAR_WIDTH)
        s.horiz_space_for_button(width)        
        x = s.x() - width
        return Number(label, event, x, s.y(), width, s.button_height, initial, min, max, tooltip)
        
    #-------------------------------------------------------------------------------
    def horiz_packed_button(s, label, event, tooltip):
        # Make space on row for button:
        width = string_width(label) + LABEL_X_BORDER
        s.horiz_space_for_button(width)
        x = s.x() - width
        return Button(label, event, x, s.y(), width, s.button_height, tooltip)

    #-------------------------------------------------------------------------------
    #-------------------------------------------------------------------------------

#-------------------------------------------------------------------------------
# Private: (not for user code)
        
    #-------------------------------------------------------------------------------
    #-------------------------------------------------------------------------------
    # Spacing between and for widgets:

    #-------------------------------------------------------------------------------
    def space_before_text(s):
        s.top_offset = s.top_offset - s.text_spacing

    #-------------------------------------------------------------------------------
    def space_for_text(s):
        s.top_offset = s.top_offset - s.text_height

    #-------------------------------------------------------------------------------
    def space_after_text(s):
        s.top_offset = s.top_offset - s.text_spacing
        
    #-------------------------------------------------------------------------------
    def space_for_button(s):
        s.top_offset = s.top_offset - s.button_height
        
    #-------------------------------------------------------------------------------
    def space_for_button_or_text(s):
        #if s.button_height > s.text_height:
            s.space_for_button()
        #else:
            #s.space_for_text()

    #-------------------------------------------------------------------------------
    def horiz_space_for_button(s, width):
        if width + s.x() > s.left + s.width:
            s.space_for_button()
            s.x_offset = 0
        s.add_to_horiz_offset(width)

    #-------------------------------------------------------------------------------
    #-------------------------------------------------------------------------------
    #-------------------------------------------------------------------------------
    #-------------------------------------------------------------------------------
    #-------------------------------------------------------------------------------

    
    #-------------------------------------------------------------------------------
    def y(s):
        return s.base_y + s.top_offset

    #-------------------------------------------------------------------------------
    def window_y(s):
        return pane_y_to_window(s.base_y + s.top_offset)

    #-------------------------------------------------------------------------------
    def x(s):
        return s.left + s.x_offset

    #-------------------------------------------------------------------------------
    def add_to_horiz_offset(s, width):
        s.x_offset = s.x_offset + width

Register(display, event, bevent)

# To create the widgets in the global dictionary before they are accessed from the
# button event callback, call the display function now (even if this is not a problem
# now, it might present as a bug later when the API is changed so do it anyway):
Redraw(0)

print 'End of script reached.'



